# ___ Code used for simplified model "Asset Price Dynamics with Limited Attention" by Terrence Hendershott, Albert J. Menkveld, Mark Seasholes, and Remy Praz ___
#
# Prelmininaries:
# - The code uses Hungarian notation meaning that the first letter(s) indicates datatype (e.g., b_xyz refers to variable xyz that is a boolean variable, df_ refers to pandas dataframe, s_ to pandas series or to python string, d_ to python double or dictionary, i_ to integer, t_ is tuple, etc.)
# - The code is written for Python 3.  It runs on OSX with Python 3.8 with all packages installed with homebrew
# - Clarifying comments are scattered throughout
# - Variable names leave out vowels except for first letter, this keeps variable names short yet makes them informative
# - The main directory contains a readme.pdf with further context and the equations pertaining to the simplified statespace model.

import time
start_time = time.time()
from IPython import embed
import pandas as pd
import numpy as np
import statsmodels.api as sm
import matplotlib.pyplot as plt
from matplotlib import rcParams
rcParams['font.sans-serif'] = "arial"; rcParams['font.family'] = "sans-serif"; rcParams.update({'mathtext.default': 'regular'})
import scipy.linalg as linalg
np.random.seed(9999) # Comment out if you want to generate a new (simulated) sample.
import scipy.optimize as optimize
pd.options.display.float_format = '{:.5f}'.format
plt.rcParams.update({'figure.figsize':(5,5), 'figure.dpi':100})

# ___ First define constants ___
i_nmbr_dys_in_smpl = 10**3 # Number of days in the sample.
d_lwr_bnd_m = 10**(-2) # Used as lower bound for riskmass to avoid "non-positive definite error forecast covariance matrix."
d_lwr_bnd_bt_M = 10**(-4) # Used for beta M, avoid "non-positive definite error forecast covariance matrix."
d_lwr_bnd_rh = -0.80 # Lower bound for correlation rho to avoid "non-positive definite error forecast covariance matrix."
d_uppr_bnd_rh = 1 - (d_lwr_bnd_rh + 1) # Upper bound for rho.
l_nms_prmtrs = ['m_i', 'm_r', 'bt_M', 'bt_w', 'sgm_w', 'rh'] # The vector theta that consists of all model parameters, investor type is in {i=institution, r=retail}.
l_nms_Y = ['G_i', 'G_r', 'MMInv', 'RetFlow', 'Return'] # The vector Y that consists of all model variables.
l_nms_cvrnc_epsln = ['epsln_11', 'epsln_12', 'espln_2', 'epsln_31', 'epsln_32'] # Names for the innovation terms.
d_lmbd = 1/20 # Lambdas captures the intensity for the slow types (i.e., institutional and retail).
d_rf = 0.0002 # Riskfree discount rate taken from Ken French' site 0.0002 ~ 1.05**(1/250).
d_bx_stl = {'facecolor':'gray', 'alpha':0.1, 'edgecolor':'white', 'boxstyle':'round,pad=.7'}
s_clr_lght_ajm = '#deebf7'; s_clr_mddl_ajm = '#9ecae1'; s_clr_drk_ajm = '#3182bd' # Color scheme used.

class htss1_sttspc_mdl(sm.tsa.statespace.MLEModel): # Create class in statsmodels to benefit from statespace model package routines, for manual see:
    # - https://www.statsmodels.org/stable/statespace.html
    # - https://www.statsmodels.org/stable/generated/statsmodels.tsa.statespace.mlemodel.MLEModel.fit.html#statsmodels.tsa.statespace.mlemodel.MLEModel.fit
    def __init__(self, endog):
        # Initialize the state space model.
        super(htss1_sttspc_mdl, self).__init__(endog, k_states=5, k_posdef=5, initialization='stationary')
        # Setup the various components of the state space representation.
        self['design'] = pd.DataFrame(np.hstack((np.full((3,2), 0), np.eye(3))), index=l_nms_Y[-3:], columns=l_nms_Y)
        self['obs_intercept'] = pd.Series(np.full((3), 0), index=l_nms_Y[-3:])
        self['state_intercept'] = pd.Series(np.full((5), 0), index=l_nms_Y)
    # Describe how parameters enter the model.
    def update(self, params, **kwargs):
        df_V, df_W, df_cvrnc_epsln = cmpt_sttspc_objcts(pd.Series(params, index=l_nms_prmtrs))
        self['transition'] = df_V
        # self['selection'] = pd.DataFrame(np.eye(5), index=l_nms_Y, columns=l_nms_Y)
        self['selection'] = df_W
        self['state_cov'] = pd.DataFrame(df_cvrnc_epsln) # This trick is needed as MLEModel does not allow the rank of the covariance matrix of espsilon to be larger than the number of states, note that df_cvrnc_epsln is 5x5 and df_W is 5x5.  The "structural" covariance matrix for epsilon is converted to a "reduced-form" one in a straightforward manner.

def cmpt_autcrsscrrltn(s_srs_x, s_srs_y, i_lg_lwr_bnd, i_lg_uppr_bnd): # This function computes autocross correlation between s_srx_x and s_srs_y wit the latter lagged between i_lg_lwr_bnd and i_lg_uppr_bnd. This correlation is an average in order to draw an observed auto-cross correlation plot that is smooth and thus readable. In other words, rather than plot a single auto-cross correlation, we plot an average auto-cross correlation across an interval of lags. The function also outputs its standard error.
    s_srs_x_stndrdzd = (s_srs_x - s_srs_x.mean())/s_srs_x.std()
    s_srs_y_stndrdzd = (s_srs_y - s_srs_y.mean())/s_srs_y.std()
    df_srs_y_stndrdzd = pd.DataFrame(s_srs_y_stndrdzd.shift(i_lg_lwr_bnd))
    for i_lg in range(i_lg_lwr_bnd+1, i_lg_uppr_bnd+1):
        df_srs_y_stndrdzd = pd.concat((df_srs_y_stndrdzd, s_srs_y_stndrdzd.shift(i_lg)), axis=1)
    l_clmn_nms = []
    for i_vl in range(i_lg_uppr_bnd-i_lg_lwr_bnd+1):
        l_clmn_nms += [s_srs_x_stndrdzd.name]
    df_srs_x_stndrdzd = pd.DataFrame(np.kron(pd.DataFrame(s_srs_x_stndrdzd), np.ones((1,i_lg_uppr_bnd-i_lg_lwr_bnd+1))), index=s_srs_x_stndrdzd.index, columns=l_clmn_nms)
    df_prdct_x_stndrdzd_y_stndrdzd_lggd = pd.DataFrame(np.multiply(df_srs_x_stndrdzd,df_srs_y_stndrdzd), index=df_srs_x_stndrdzd.index)
    i_nmbr_lgs = i_lg_uppr_bnd-i_lg_lwr_bnd+1
    s_wghts = pd.Series(np.full((i_nmbr_lgs), 1/i_nmbr_lgs), index=df_prdct_x_stndrdzd_y_stndrdzd_lggd.columns.tolist(), name='wghts')
    d_autcrsscrrltn_estmtr = s_wghts.dot(df_prdct_x_stndrdzd_y_stndrdzd_lggd.mean())
    d_autcrsscrrltn_estmtr_stndrd_errr = np.sqrt(np.linalg.multi_dot([s_wghts, df_prdct_x_stndrdzd_y_stndrdzd_lggd.cov()  / df_prdct_x_stndrdzd_y_stndrdzd_lggd.count(), s_wghts]))
    return d_autcrsscrrltn_estmtr, d_autcrsscrrltn_estmtr_stndrd_errr

def cmpt_emprcl_vrnc_autcrsscrrltn_Y(df_smpl): # This function uses the function cmpt_autcrsscrrltn to compute the observed variance and auto-cross correlations for all three empirical variables: MMInv, RetFlow, and Return.
    l_nms_vrbls = df_smpl.columns.tolist()
    # ___ First variance (with standard error). ___
    l_emprcl_vrnc_estmtr = []
    l_emprcl_vrnc_estmtr_stndrd_err = []
    for s_clmn, s_srs in df_smpl.iteritems(): # Iterate over the three empirical variables.
        s_srs_dmnd_sqrd = (s_srs - s_srs.mean()).apply(lambda x: x**2) # Note that dmnd abbreviates demeaned.
        l_emprcl_vrnc_estmtr = l_emprcl_vrnc_estmtr + [s_srs_dmnd_sqrd.mean()]
        l_emprcl_vrnc_estmtr_stndrd_err = l_emprcl_vrnc_estmtr_stndrd_err + [s_srs_dmnd_sqrd.std() / np.sqrt(s_srs_dmnd_sqrd.count())] # This is the standard error estimate for the variance estimator (this is easily seen by considering Y=(X-mean(X))**2 as an estimator for the variance, the standard error of Y is then sqrt(var(Y)/count(Y)).
    s_emprcl_vrnc_estmtr = pd.Series(l_emprcl_vrnc_estmtr, index=l_nms_vrbls)
    s_emprcl_vrnc_estmtr_stndrd_errr = pd.Series(l_emprcl_vrnc_estmtr_stndrd_err, index=l_nms_vrbls)
    # ___ then autocrosscorrelations (with standard error) ___
    df_lgs = pd.DataFrame([[0,0], [1,1], [2,4], [5,9], [10,14], [15,19]], columns=['i_lg_lwr_bnd','i_lg_uppr_bnd'], index=pd.Series([0, 1, 3, 8, 13, 18], name='i_lg')) # These are the lag intervals over which the autocorrelations are computed.
    df_emprcl_autcrsscrrltn = pd.DataFrame( np.full((df_lgs.shape[0],9), np.nan), index=df_lgs.index, columns=pd.MultiIndex.from_product([l_nms_vrbls, l_nms_vrbls], names=['frst_vrbl', 'scnd_vrbl_lggd']) )
    df_emprcl_autcrsscrrltn.index.name = 'lg'
    df_emprcl_autcrsscrrltn_stndrd_errr = np.nan * df_emprcl_autcrsscrrltn
    for s_frst_vrbl in l_nms_vrbls:
        for s_scnd_vrbl in l_nms_vrbls:
            for i_lg, s_lg_bnds in df_lgs.iterrows():
                d_autcrsscrrltn, d_autcrsscrrltn_stndrd_errr = cmpt_autcrsscrrltn(**{'s_srs_x':df_smpl[s_frst_vrbl], 's_srs_y':df_smpl[s_scnd_vrbl], 'i_lg_lwr_bnd':s_lg_bnds.i_lg_lwr_bnd, 'i_lg_uppr_bnd':s_lg_bnds.i_lg_uppr_bnd})
                df_emprcl_autcrsscrrltn.loc[i_lg][s_frst_vrbl, s_scnd_vrbl] = d_autcrsscrrltn
                df_emprcl_autcrsscrrltn_stndrd_errr.loc[i_lg][s_frst_vrbl, s_scnd_vrbl] = d_autcrsscrrltn_stndrd_errr
    return s_emprcl_vrnc_estmtr, s_emprcl_vrnc_estmtr_stndrd_errr, df_emprcl_autcrsscrrltn, df_emprcl_autcrsscrrltn_stndrd_errr

def cmpt_sttspc_objcts(s_prmtrs): # This function computes all statespace objects based on the input parameters.
    # To keep code readable a new pointer to s_prmtrs is created below.
    s_p = s_prmtrs

    # ___ V (matrix statespace model). ___
    df_V = pd.DataFrame([
        [np.exp(-d_lmbd)                          , 0                                        , 0, 0, 0],
        [0                                        , np.exp(-d_lmbd)                          , 0, 0, 0],
        [s_p.bt_M*np.exp(-d_lmbd)                 , s_p.bt_M*np.exp(-d_lmbd)                 , 0, 0, 0],
        [0                                        , 1-np.exp(-d_lmbd)                        , 0, 0, 0],
        [s_p.bt_w*(1-np.exp(-d_lmbd))/(d_rf+d_lmbd), s_p.bt_w*(1-np.exp(-d_lmbd))/(d_rf+d_lmbd), 0, 0, 0],
        ], index=l_nms_Y, columns=l_nms_Y)

    # ___ W (selection matrix statespace model). ___
    df_W = pd.DataFrame([
        [1                        , 0                        , 0, 0, 0],
        [0                        , 1                        , 0, 0, 0],
        [s_p.bt_M                 , s_p.bt_M                 , 0, 0, 0],
        [0                        , -1                       , 0, 0, 1],
        [-s_p.bt_w/(d_rf + d_lmbd), -s_p.bt_w/(d_rf + d_lmbd), 1, 0, 0]
        ], index=l_nms_Y, columns=l_nms_cvrnc_epsln)


    # ___ Covariance epsilon. ___
    d_A = (1-np.exp(-2*d_lmbd)) / (2*d_lmbd)
    d_B = (1-np.exp(-  d_lmbd)) / (  d_lmbd)
    df_cvrnc_epsln = pd.DataFrame([
        [(s_p.m_i**2)*d_A               , (s_p.rh**2)*s_p.m_i*s_p.m_r*d_A, s_p.rh*s_p.m_i*s_p.sgm_w*d_B          , (s_p.m_i**2)*d_B               , (s_p.rh**2)*s_p.m_i*s_p.m_r*d_B],
        [(s_p.rh**2)*s_p.m_i*s_p.m_r*d_A, (s_p.m_r**2)*d_A               , s_p.rh*s_p.m_r*s_p.sgm_w*d_B          , (s_p.rh**2)*s_p.m_i*s_p.m_r*d_B, (s_p.m_r**2)*d_B],
        [s_p.rh*s_p.m_i*s_p.sgm_w*d_B   , s_p.rh*s_p.m_r*s_p.sgm_w*d_B   , s_p.sgm_w**2                          , s_p.rh*s_p.m_i*s_p.sgm_w       , s_p.rh*s_p.m_r*s_p.sgm_w],
        [(s_p.m_i**2)*d_B               , (s_p.rh**2)*s_p.m_i*s_p.m_r*d_B, s_p.rh*s_p.m_i*s_p.sgm_w              , s_p.m_i**2                     , (s_p.rh**2)*s_p.m_i*s_p.m_r],
        [(s_p.rh**2)*s_p.m_i*s_p.m_r*d_B, (s_p.m_r**2)*d_B               , s_p.rh*s_p.m_r*s_p.sgm_w              , (s_p.rh**2)*s_p.m_i*s_p.m_r    , s_p.m_r**2],
        ], index=l_nms_cvrnc_epsln, columns=l_nms_cvrnc_epsln)
    return df_V, df_W, df_cvrnc_epsln

def cmpt_mdl_impld_vrnc_autcrsscrrltn_Y(s_prmtrs, i_mxm_lg): # This function computes the model-implied variance and auto-cross correlations up to lag i_mxm_lg for all model variables (collected in Y).

    # ___ First compute the statespace model objects given the model parameters collected in s_prmtrs. ___
    df_V, df_W, df_cvrnc_epsln = cmpt_sttspc_objcts(s_prmtrs)

    # ___ Then compute the model-implied variance for discretely sampled variables collected in Y (see equation (xyz) in the paper). ___
    df_cvrnc_G = pd.DataFrame(np.full((2,2), np.nan), index=l_nms_prmtrs[:2], columns=l_nms_prmtrs[:2])
    for s_rw in l_nms_prmtrs[:2]:
        for s_clmn in l_nms_prmtrs[:2]:
            if s_rw==s_clmn:
                df_cvrnc_G.loc[s_rw, s_clmn] = (s_prmtrs[s_rw]*s_prmtrs[s_clmn])/(2*d_lmbd)
            else:
                df_cvrnc_G.loc[s_rw, s_clmn] = ((s_prmtrs['rh']**2)*s_prmtrs[s_rw]*s_prmtrs[s_clmn])/(2*d_lmbd)
    df_V_tld = df_V.iloc[:,:2]
    df_mdl_impld_cvrnc_Y= pd.DataFrame(np.dot(df_V_tld, np.dot(df_cvrnc_G, df_V_tld.transpose())) + np.dot(df_W, np.dot(df_cvrnc_epsln, df_W.transpose())), index=l_nms_Y, columns=l_nms_Y)

    # ___ Finally compute the model-implied autocorrelations (see equation (xyz) in the paper). ___
    df_mdl_impld_autcrsscrrltn_Y = pd.DataFrame(np.full((i_mxm_lg+1, 25), np.nan), index=range(i_mxm_lg+1), columns=pd.MultiIndex.from_product([l_nms_Y, l_nms_Y], names=['frst_vrbl', 'scnd_vrbl_lggd']) )
    df_mdl_impld_autcrsscrrltn_Y.index.name = 'lg'
    for i_lg in range(0,i_mxm_lg+1):
        if i_lg==0:
            df_cvrnc_Y_t_Y_t_mns_i_lg = df_mdl_impld_cvrnc_Y
        else:
            df_cvrnc_Y_t_Y_t_mns_i_lg = pd.DataFrame(np.linalg.multi_dot( [df_V.iloc[:,:2], linalg.expm(-(i_lg-1)*(d_lmbd*np.eye(2))), df_mdl_impld_cvrnc_Y.iloc[:,:2].transpose()] ), index=l_nms_Y, columns=l_nms_Y)
        for s_frst_vrbl in l_nms_Y: # Turn autocovariance to autocorrelation.
            for s_scnd_vrbl in l_nms_Y:
                d_vrnc_x_tms_vrnc_y = df_mdl_impld_cvrnc_Y[s_frst_vrbl][s_frst_vrbl]*df_mdl_impld_cvrnc_Y[s_scnd_vrbl][s_scnd_vrbl]
                if  d_vrnc_x_tms_vrnc_y>0: # Otherwise division by zero.
                    df_mdl_impld_autcrsscrrltn_Y.loc[i_lg][s_frst_vrbl, s_scnd_vrbl] = (df_cvrnc_Y_t_Y_t_mns_i_lg.loc[s_frst_vrbl][s_scnd_vrbl]) / np.sqrt(d_vrnc_x_tms_vrnc_y)
                else:
                    df_mdl_impld_autcrsscrrltn_Y.loc[i_lg][s_frst_vrbl, s_scnd_vrbl] = np.nan
    return df_mdl_impld_cvrnc_Y, df_mdl_impld_autcrsscrrltn_Y

def cmpt_strtng_vls(df_dt_smpl): # Find reasonable starting value for the maximum-likelihood optimization.  The formulas and their motivation is in Appendix XYZ of the paper.
    s_strtng_vls_prmtrs = pd.Series((d_lwr_bnd_m, d_lwr_bnd_m, d_lwr_bnd_bt_M, 0, 0, 0), index=l_nms_prmtrs, name='strtng_vls') # Initialize the parameter vector.
    d_m_r_sqrd = df_dt_smpl.RetFlow.cov(df_dt_smpl.RetFlow.shift(20)) / ( (1-np.exp(-d_lmbd)) * np.exp(-(20-1)*d_lmbd) * (np.exp(-d_lmbd)*(1-np.exp(-d_lmbd))/(2*d_lmbd) + (1-np.exp(-d_lmbd))/d_lmbd - (1-np.exp(-2*d_lmbd))/(2*d_lmbd)) ) # Calibrate the risk-mass of retail investors to autocovariance in retail flow at lag 20 (could be other lags but this nicely corresponds to baseline arrival intensity of 1/20.
    if d_m_r_sqrd>d_lwr_bnd_m: # Only keep the "calibrated" value if it exceeds the lower bound (d_lwr_bnd_m is a very small number to avoid non-positive definite prediction error variance matrix in the optimization procedure).
        s_strtng_vls_prmtrs['m_r'] = np.sqrt(d_m_r_sqrd)
    d_bt_M = df_dt_smpl.RetFlow.cov(df_dt_smpl.MMInv.shift(1)) / ( (1-np.exp(-d_lmbd)) * (s_strtng_vls_prmtrs.m_r**2)/(2*d_lmbd) ) # Calibrate the risk-mass of retail investors to cross-autocovariance RetFlow and MMInv.
    if (d_bt_M>d_lwr_bnd_bt_M)&(not(np.isinf(d_bt_M))): # If beta M exceeds its lower bound and is finite, then case change the initialization value of beta_M to d_bt_M and continue to find starting values for risk-mass slow institutions and beta_W.
        s_strtng_vls_prmtrs.bt_M = d_bt_M
        d_m_all_sqrd = df_dt_smpl.MMInv.cov(df_dt_smpl.MMInv.shift(20)) / ( (s_strtng_vls_prmtrs.bt_M**2) * np.exp(-20*d_lmbd) / (2*d_lmbd) )
        d_m_i_sqrd = d_m_all_sqrd - s_strtng_vls_prmtrs.m_r**2
        if d_m_i_sqrd > 0: # Change default value of zero.
            s_strtng_vls_prmtrs.m_i = np.sqrt(d_m_i_sqrd)
        d_bt_w = df_dt_smpl.Return.cov(df_dt_smpl.MMInv.shift(1)) / ( s_strtng_vls_prmtrs.bt_M * ((1-np.exp(-d_lmbd))/(d_rf + d_lmbd)) * ((s_strtng_vls_prmtrs.m_i**2+s_strtng_vls_prmtrs.m_r**2)/(2*d_lmbd)) )
        if (d_bt_w>0)&(not(np.isinf(d_bt_w))): # Change default value of zero.
            s_strtng_vls_prmtrs.bt_w = d_bt_w
    d_sgm_w_sqrd = df_dt_smpl.Return.var() - (s_strtng_vls_prmtrs.bt_w**2) *  ( (1-np.exp(-d_lmbd))/(d_lmbd*((d_rf + d_lmbd)**2)) * (s_strtng_vls_prmtrs.m_i**2+s_strtng_vls_prmtrs.m_r**2) )
    if d_sgm_w_sqrd>0:
        s_strtng_vls_prmtrs.sgm_w = np.sqrt(d_sgm_w_sqrd)
    df_mdl_impld_cvrnc_Y_at_strtng_vls, df_mdl_impld_autcrsscrrltn_Y_at_strtng_vls = cmpt_mdl_impld_vrnc_autcrsscrrltn_Y(s_strtng_vls_prmtrs, 0)
    d_excss_mrktmkrinvntry_rtrn_cvrnc = df_dt_smpl.MMInv.cov(df_dt_smpl.Return) - df_mdl_impld_cvrnc_Y_at_strtng_vls.loc['MMInv','Return']
    d_rh = d_excss_mrktmkrinvntry_rtrn_cvrnc / (np.sqrt(df_mdl_impld_cvrnc_Y_at_strtng_vls.loc['MMInv','MMInv'])*s_strtng_vls_prmtrs.sgm_w)
    if -1 < d_rh < 1: # Change correlation between fundamental-value innovations (i.e., dividend), rho, from its default value of zero to match "excess" covariance between market maker inventory and return.
        s_strtng_vls_prmtrs.rh = d_rh
    return s_strtng_vls_prmtrs

def plt_autcvrnc(s_emprcl_vrnc_estmtr, df_emprcl_autcrsscrrltn, s_prmtrs, df_mdl_impld_cvrnc_Y, df_mdl_impld_autcrsscrrltn_Y): # Signature plot for three observed model variables (MMInv, RetFlow, Return).  Cross-autocorrelation plots up to lag 20 (approximately one month) and the unconditional standard deviations (see Figure XYZ in paper).  It is based on observed as well as model-implied values.
    f_autcvrnc, a_autcvrnc = plt.subplots(4, 3, figsize=list(1.5*np.array((6,6))))
    f_autcvrnc.suptitle('Model: $\mu_i$ = %.1f, $\mu_r$ = %.1f, $\\beta_M$ = %.4f, $\\beta_w$ = %.4f, $\sigma_w$ = %.0f, $\\rho$ = %.2f'%(s_prmtrs.m_i, s_prmtrs.m_r, s_prmtrs.bt_M, s_prmtrs.bt_w, s_prmtrs.sgm_w, s_prmtrs.rh), bbox=d_bx_stl, fontsize='medium') # Model parameters on which model-implied values plots are based are put in the header of the figure.
    for i_rw, s_rw_vrbl_nm  in enumerate(l_nms_Y[-3:]): # The top part of the figure contains 3x3 plots of cross-autocorrelations.
        for i_clmn, s_clmn_vrbl_nm  in enumerate(l_nms_Y[-3:]):
            a_autcvrnc[i_rw, i_clmn].plot(df_emprcl_autcrsscrrltn[s_clmn_vrbl_nm, s_rw_vrbl_nm], label='Observed', markerfacecolor=s_clr_drk_ajm, marker='o', ls='None') # Row and column are changed because we need to read off a lead correlation from an object that works with lags.
            a_autcvrnc[i_rw, i_clmn].plot(df_mdl_impld_autcrsscrrltn_Y[s_clmn_vrbl_nm, s_rw_vrbl_nm], label='Model-implied', c=s_clr_mddl_ajm, lw=2)
            a_autcvrnc[i_rw, i_clmn].axhline(y=0, c='k', lw=1)
            a_autcvrnc[i_rw, i_clmn].set_xlabel('Lead k (in days)')
            a_autcvrnc[i_rw, i_clmn].set_ylabel('$Cor(%s_t,%s_{t+k})$'%(s_rw_vrbl_nm, s_clmn_vrbl_nm))
            if (i_rw==0) and (i_clmn==0):
                a_autcvrnc[i_rw, i_clmn].legend(loc='upper right')
            if (i_rw==2) and (i_clmn==2):
                a_autcvrnc[i_rw, i_clmn].set_ylim(top=0.010, bottom=-0.030)
    for i_clmn, s_clmn_vrbl_nm  in enumerate(l_nms_Y[-3:]): # The bottom part of the figure contains a row with the (unconditional) standard deviation of the variables.
        d_unts_vrbls = {'MMInv':'Million \$', 'RetFlow':'Million \$', 'Return':'Perunage'}
        a_autcvrnc[3, i_clmn].bar([1], np.sqrt(df_mdl_impld_cvrnc_Y[s_clmn_vrbl_nm][s_clmn_vrbl_nm]), alpha=0.2, color=s_clr_mddl_ajm)
        a_autcvrnc[3, i_clmn].plot([1], np.sqrt(s_emprcl_vrnc_estmtr[s_clmn_vrbl_nm]), markerfacecolor=s_clr_drk_ajm, marker='o')
        a_autcvrnc[3, i_clmn].set_ylabel('%s (%s)'%(s_clmn_vrbl_nm, d_unts_vrbls[s_clmn_vrbl_nm]))
        a_autcvrnc[3, i_clmn].set_xticks([])
        t_ylm = a_autcvrnc[3, i_clmn].get_ylim()
        a_autcvrnc[3, i_clmn].set_ylim(t_ylm[0], 1.1*t_ylm[1])
    plt.tight_layout(rect=[0, 0, 1, 0.925])
    return f_autcvrnc, a_autcvrnc

# ___ Begin main program ___

# ___ Prepare data, either simulate the data or take them from the input directory. ___
d_tr_prmtrs = {'m_i':25, 'm_r':5, 'bt_M':0.01, 'bt_w':0.10, 'sgm_w':200, 'rh':-0.25}
s_tr_prmtrs = pd.Series(d_tr_prmtrs, name='tr_prmtrs') # These are the true model parameters from which the data will be generated and put into the input directory.
sm_sttspc_mdl = htss1_sttspc_mdl(np.full((1,3), None)) # Initialize model.
df_smpl = pd.DataFrame(sm_sttspc_mdl.simulate(params=s_tr_prmtrs, nsimulations=i_nmbr_dys_in_smpl), columns=l_nms_Y[-3:]) # Simulate
df_smpl[['MMInv','RetFlow','Return']].plot(subplots=True, color=[s_clr_mddl_ajm, s_clr_mddl_ajm, s_clr_mddl_ajm])
plt.savefig('../output/all-inpt-srs.pdf') # Show the simulated series.
df_smpl.to_csv('../input/smltd_srs.csv')

# ___ Compute empirical objects ___
s_emprcl_vrnc_estmtr, s_emprcl_vrnc_estmtr_stndrd_errr, df_emprcl_autcrsscrrltn, df_emprcl_autcrsscrrltn_stndrd_errr = cmpt_emprcl_vrnc_autcrsscrrltn_Y(df_smpl[['MMInv','RetFlow','Return']])

# ___ Find reasonable starting values to start maximum-likelihood optimization (see Appendix XYZ in paper). ___
s_strtng_vls_prmtrs = cmpt_strtng_vls(df_smpl)
df_mdl_impld_cvrnc_Y_at_strtng_vls, df_mdl_impld_autcrsscrrltn_Y_at_strtng_vls = cmpt_mdl_impld_vrnc_autcrsscrrltn_Y(s_strtng_vls_prmtrs, 20)

# ___ Estimate the statespace model with maximum likelihood. ___
sm_smpl = htss1_sttspc_mdl(df_smpl[['MMInv','RetFlow','Return']])
t_bnds = ((d_lwr_bnd_m, None), (d_lwr_bnd_m, None), (d_lwr_bnd_bt_M, 1), (0, None), (0, None), (d_lwr_bnd_rh, d_uppr_bnd_rh)) # tuple with variable bounds
sm_smpl._param_names = l_nms_prmtrs
sm_smpl._state_names = l_nms_Y
sm_estmtn_rslt = sm_smpl.fit(start_params=s_strtng_vls_prmtrs.to_numpy(), bounds=t_bnds, maxiter=100) # For more options that could be set see https://docs.w3cub.com/statsmodels/generated/statsmodels.discrete.discrete_model.binarymodel.fit/.
s_prmtr_estmts = sm_estmtn_rslt.params; s_prmtr_estmts.name = 'prmtr_estmts'
s_prmtr_stndrd_errrs = sm_estmtn_rslt.bse; s_prmtr_stndrd_errrs.name = 'stndrd_errrs'
print('\n---\nThe true parameters, starting values, and parameter estimates and their standard errors are:\n---\n', pd.concat([s_tr_prmtrs, s_strtng_vls_prmtrs, s_prmtr_estmts, s_prmtr_stndrd_errrs], axis=1), '\n---\n')

# ___ Print signature plot paper (Figure XYZ)___
df_mdl_impld_cvrnc_Y, df_mdl_impld_autcrsscrrltn_Y = cmpt_mdl_impld_vrnc_autcrsscrrltn_Y(s_prmtr_estmts, 20)
f_autcvrnc, a_autcvrnc = plt_autcvrnc(s_emprcl_vrnc_estmtr, df_emprcl_autcrsscrrltn, s_prmtr_estmts, df_mdl_impld_cvrnc_Y, df_mdl_impld_autcrsscrrltn_Y)
plt.savefig('../output/autcvrnc_prmtr_estmts.pdf'); plt.close()

d_m, d_s = divmod(time.time()-start_time, 60); d_h, d_m = divmod(d_m, 60)
print("---\nDone! It all took %d:%02d:%02d (h:m:s) to finish.\n---"%(d_h, d_m, d_s))
